/************************************************************
 **
 ** pic12.c : 12-bit pic code generation definitions
 **
 ** Copyright (c) 2007, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include "pic12.h"

boolean_t pic12_code_to_pcode(pfile_t *pf, pic_code_t code,
    unsigned val, unsigned literal, pic_code_to_pcode_t *pdst)
{
  unsigned dst;
  unsigned bit;
  unsigned lit8;
  unsigned lit9; /* branch literal */
  unsigned pcode;

  bit   = literal & 0x07;
  lit8  = literal & 0x00ff; /* 8 bit literal */
  lit9  = literal & 0x01ff; /* 9 bit literal (goto) */
  dst   = (pic_opdst_f == pic_code_dst_get(code)) << 5;
  pcode = 0xffff; /* an impossible code (it's greater than 14 bits) */
  switch (pic_code_op_get(code)) {
    /* directives (no code) */
    case pic_opcode_org:    /* ORG xxxx */
    case pic_opcode_end:    /* END      */
    case pic_opcode_none:   /* no code here */
    case pic_opcode_branchlo_nop:
    case pic_opcode_branchhi_nop:
      break;
    case pic_opcode_nop:    pcode = 0x0000; break;
    /* file operator (code, dst, val) */
    case pic_opcode_movwf:  pcode = 0x0000 | 0x20 | val; break;
    case pic_opcode_clrf:   pcode = 0x0000 | 0x60 | val; break;
    case pic_opcode_subwf:  pcode = 0x0080 |  dst | val; break;
    case pic_opcode_decf:   pcode = 0x00c0 |  dst | val; break;
    case pic_opcode_iorwf:  pcode = 0x0100 |  dst | val; break;
    case pic_opcode_andwf:  pcode = 0x0140 |  dst | val; break;
    case pic_opcode_xorwf:  pcode = 0x0180 |  dst | val; break;
    case pic_opcode_addwf:  pcode = 0x01c0 |  dst | val; break;
    case pic_opcode_movf:   pcode = 0x0200 |  dst | val; break;
    case pic_opcode_comf:   pcode = 0x0240 |  dst | val; break;
    case pic_opcode_incf:   pcode = 0x0280 |  dst | val; break;
    case pic_opcode_decfsz: pcode = 0x02c0 |  dst | val; break;
    case pic_opcode_rrcf:
    case pic_opcode_rrf:    pcode = 0x0300 |  dst | val; break;
    case pic_opcode_rlcf:
    case pic_opcode_rlf:    pcode = 0x0340 |  dst | val; break;
    case pic_opcode_swapf:  pcode = 0x0380 |  dst | val; break;
    case pic_opcode_incfsz: pcode = 0x03c0 |  dst | val; break;
    /* simple opcodes (no operands) */
    case pic_opcode_retfie: pcode = 0xffff; break;
    case pic_opcode_return: pcode = 0xffff; break;
    case pic_opcode_sleep:  pcode = 0x0003; break;
    case pic_opcode_clrwdt: pcode = 0x0004; break;
    case pic_opcode_clrw:   pcode = 0x0040; break;
    /* bit oriented (3 bit literal) */
    case pic_opcode_bcf:    pcode = 0x0400 | (bit << 5) | val; break;
    case pic_opcode_bsf:    pcode = 0x0500 | (bit << 5) | val; break;
    case pic_opcode_btfsc:  pcode = 0x0600 | (bit << 5) | val; break;
    case pic_opcode_btfss:  pcode = 0x0700 | (bit << 5) | val; break;
    /* literal & w  (8 bit literal) */
    case pic_opcode_movlw:  pcode = 0x0c00 | lit8; break;
    case pic_opcode_retlw:  pcode = 0x0800 | lit8; break;
    case pic_opcode_iorlw:  pcode = 0x0d00 | lit8; break;
    case pic_opcode_andlw:  pcode = 0x0e00 | lit8; break;
    case pic_opcode_xorlw:  pcode = 0x0f00 | lit8; break;
    case pic_opcode_sublw:
    case pic_opcode_addlw:
       pfile_log(pf, pfile_log_err, "%slw not supported!",
           (pic_opcode_sublw == pic_code_op_get(code)) ? "sub" : "add");
       break;
    /* branching (8 or 9 bit literal) */
    case pic_opcode_call:   
       if (lit9 > 256) {
         pfile_log(pf, pfile_log_err, "call past 0x1ff");
       }
       pcode = 0x0900 | lit8; 
       break;
    case pic_opcode_goto:   pcode = 0x0a00 | lit9; break;
    /* special (pseudo) opcodes */
    case pic_opcode_datalo_set: /* bsf _fsr, 5 */
      pcode = 0x0500U | (5 << 5) | 0x0004;
      break;
    case pic_opcode_datalo_clr: /* bcf _fsr, 5 */
      pcode = 0x0400U | (5 << 5) | 0x0004;
      break;
    case pic_opcode_datahi_set: /* bsf _fsr, 6 */
      pcode = 0x0500U | (6 << 5) | 0x0004;
      break;
    case pic_opcode_datahi_clr: /* bcf _fsr, 6 */
      pcode = 0x0400U | (6 << 5) | 0x0004;
      break;
    case pic_opcode_irp_set:
    case pic_opcode_irp_clr:
      pfile_log(pf, pfile_log_err, "irp macros not supported");
      break;
    case pic_opcode_branchlo_set: /* bsf _status, 5 */
      pcode = 0x0500U | (5 << 5) | 0x0003;
      break;
    case pic_opcode_branchlo_clr: /* bcf _status, 5 */
      pcode = 0x0400U | (5 << 5) | 0x0003;
      break;
    case pic_opcode_branchhi_set: /* bsf _status, 6 */
      pcode = 0x0500U | (6 << 5) | 0x0003;
      break;
    case pic_opcode_branchhi_clr: /* bcf _status, 6 */
      pcode = 0x0400U | (6 << 5) | 0x0003;
      break;
    case pic_opcode_option:       /* option */
      pcode = 0x0002U;
      break;
    case pic_opcode_tris:         /* tris 6 or 7 */
      pcode = lit8 & 0x07;
      break;
    case pic_opcode_db: /* this is handled elsewhere */
      break;
    case pic_opcode_mullw:
    case pic_opcode_addwfc:
    case pic_opcode_dcfsnz:
    case pic_opcode_infsnz:
    case pic_opcode_rlncf:
    case pic_opcode_rrncf:
    case pic_opcode_subfwb:
    case pic_opcode_subwfb:
    case pic_opcode_cpfseq:
    case pic_opcode_cpfsgt:
    case pic_opcode_cpfslt:
    case pic_opcode_mulwf:
    case pic_opcode_negf:
    case pic_opcode_setf:
    case pic_opcode_tstfsz:
    case pic_opcode_btg:
    case pic_opcode_bc:
    case pic_opcode_bn:
    case pic_opcode_bnc:
    case pic_opcode_bnn:
    case pic_opcode_bnov:
    case pic_opcode_bnz:
    case pic_opcode_bov:
    case pic_opcode_bz:
    case pic_opcode_bra:
    case pic_opcode_rcall:
    case pic_opcode_daw:
    case pic_opcode_pop:
    case pic_opcode_push:
    case pic_opcode_reset:
    case pic_opcode_lfsr:
    case pic_opcode_movff:
    case pic_opcode_movlb:
    case pic_opcode_movlp:
    case pic_opcode_tblrd:
    case pic_opcode_tblwt:
      pfile_log(pf, pfile_log_err, "illegal instruction");
      break;
  }
  pdst->ct = 0;
  if (0xffffU != pcode) {
    pdst->ct      = 1;
    pdst->code[0] = pcode;
  }
  return pdst->ct != 0;
}

void pic12_asm_header_write(pfile_t *pf, variable_const_t n_code_sz)
{
  pfile_write(pf, pfile_write_asm, 
      "datahi_set macro val\n"
      "  bsf 4, 6 ; FSR<6>\n"
      "  endm\n");
  pfile_write(pf, pfile_write_asm,
      "datahi_clr macro val\n"
      "  bcf 4, 6 ; FSR<6>\n"
      "  endm\n");
  pfile_write(pf, pfile_write_asm,
      "datalo_set macro val\n"
      "  bsf 4, 5 ; FSR<5>\n"
      "  endm\n");
  pfile_write(pf, pfile_write_asm,
      "datalo_clr macro val\n"
      "  bcf 4, 5 ; FSR<5>\n"
      "  endm\n");
  pfile_write(pf, pfile_write_asm,
      "branchhi_set macro lbl\n"
      "  %s\n"
      "  endm\n",
      (n_code_sz < 1024) ? "nop" : "  bsf 3, 6 ; STATUS<6>");
  pfile_write(pf, pfile_write_asm,
      "branchhi_clr macro lbl\n"
      "  %s\n"
      "  endm\n",
      (n_code_sz < 1024) ? "nop" : "  bcf 3, 6 ; STATUS<6>");
  pfile_write(pf, pfile_write_asm,
      "branchlo_set macro lbl\n"
      "  %s\n"
      "  endm\n",
      (n_code_sz  < 512) ? "nop" : "  bsf 3, 5 ; STATUS<5>");
  pfile_write(pf, pfile_write_asm,
      "branchlo_clr macro lbl\n"
      "  %s\n"
      "  endm\n",
      (n_code_sz  < 512) ? "nop" : "  bcf 3, 5 ; STATUS<5>");
  if (pfile_flag_test(pf, PFILE_FLAG_DEBUG_COMPILER)) {
    pfile_write(pf, pfile_write_asm,
        "branchhi_nop macro lbl\n"
        "  endm\n");
    pfile_write(pf, pfile_write_asm,
        "branchlo_nop macro lbl\n"
        "  endm\n");
  }
}

